/* 
   wifiboy32.c
     - Fast SPI LCD driver - spi_master driver version
     - Font driver - 
     - Sprite Engine
     - Essential Drawing library
     - Sound/SFX driver
     - Button driver
     - Timer/Ticker driver

   tklua@wifiboy.org, Oct 13th, 2017
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_system.h"
#include "driver/spi_master.h"
#include "soc/gpio_struct.h"
#include "driver/gpio.h"
#include "wifiboy32.h"
#include "wb32_fonts.h"

#define PIN_NUM_MISO 12
#define PIN_NUM_MOSI 13
#define PIN_NUM_CLK  14
#define PIN_NUM_CS   15
#define PIN_NUM_DC   2
#define PIN_NUM_BCKL 4

typedef struct {
    uint8_t cmd;
    uint8_t data[16];
    uint8_t databytes;
} lcd_init_cmd_t;

spi_device_handle_t _lcd_spi;

DRAM_ATTR static const lcd_init_cmd_t lcd_init_cmds[]={
    {0xCF, {0x00, 0x83, 0X30}, 3},
    {0xED, {0x64, 0x03, 0X12, 0X81}, 4},
    {0xE8, {0x85, 0x01, 0x79}, 3},
    {0xCB, {0x39, 0x2C, 0x00, 0x34, 0x02}, 5},
    {0xF7, {0x20}, 1},
    {0xEA, {0x00, 0x00}, 2},
    {0xC0, {0x26}, 1},
    {0xC1, {0x11}, 1},
    {0xC5, {0x35, 0x3E}, 2},
    {0xC7, {0xBE}, 1},
    {0x36, {0x48}, 1},
    {0x3A, {0x55}, 1},
    {0xB1, {0x00, 0x1B}, 2},
    {0xF2, {0x08}, 1},
    {0x26, {0x01}, 1},
    {0xE0, {0x1F, 0x1A, 0x18, 0x0A, 0x0F, 0x06, 0x45, 0X87, 0x32, 0x0A, 0x07, 0x02, 0x07, 0x05, 0x00}, 15},
    {0XE1, {0x00, 0x25, 0x27, 0x05, 0x10, 0x09, 0x3A, 0x78, 0x4D, 0x05, 0x18, 0x0D, 0x38, 0x3A, 0x1F}, 15},
    {0x2A, {0x00, 0x00, 0x00, 0xEF}, 4},
    {0x2B, {0x00, 0x00, 0x01, 0x3f}, 4}, 
    {0x2C, {0}, 0},
    {0xB7, {0x07}, 1},
    {0xB6, {0x0A, 0x82, 0x27, 0x00}, 4},
    {0x11, {0}, 0x80},
    {0x29, {0}, 0x80},
    {0, {0}, 0xff},
};

void lcd_cmd(const uint8_t cmd) 
{
    spi_transaction_t t;
    memset(&t, 0, sizeof(t));
    t.length=8;
    t.tx_buffer=&cmd;
    t.user=(void*)0;
    spi_device_transmit(_lcd_spi, &t);
}

void lcd_data(const uint8_t *data, int len) 
{
    spi_transaction_t t;
    if (len==0) return;
    memset(&t, 0, sizeof(t));
    t.length=len*8;
    t.tx_buffer=data;
    t.user=(void*)1;
    spi_device_transmit(_lcd_spi, &t);
}

void lcd_spi_pre_transfer_callback(spi_transaction_t *t) 
{
    int dc=(int)t->user;
    gpio_set_level(PIN_NUM_DC, dc);
}

void lcd_init() 
{
    int cmd=0;
    gpio_set_direction(PIN_NUM_DC, GPIO_MODE_OUTPUT);
    gpio_set_direction(PIN_NUM_BCKL, GPIO_MODE_OUTPUT);
    while (lcd_init_cmds[cmd].databytes!=0xff) {
        lcd_cmd(lcd_init_cmds[cmd].cmd);
        lcd_data(lcd_init_cmds[cmd].data, lcd_init_cmds[cmd].databytes&0x1F);
        if (lcd_init_cmds[cmd].databytes&0x80) vTaskDelay(100 / portTICK_RATE_MS);
        cmd++;
    }
    gpio_set_level(PIN_NUM_BCKL, 1);
}

static void wait_trans(int n) 
{
    spi_transaction_t *rtrans;
    for (int i=0; i<n; i++) spi_device_get_trans_result(_lcd_spi, &rtrans, portMAX_DELAY);
}

static spi_transaction_t trans[6];

void wb32_setAddrWindow(uint16_t x0, uint16_t y0, uint16_t x1, uint16_t y1) 
{
    memset(&trans[0], 0, 5*sizeof(spi_transaction_t));
    for(int x=0; x<5; x++) {
        if ((x&1)==0) {
          trans[x].length=8;
          trans[x].user=(void*)0;
        } else {
          trans[x].length=8*4;
          trans[x].user=(void*)1;
        }
        trans[x].flags=SPI_TRANS_USE_TXDATA;
    }
    trans[0].tx_data[0]=0x2A;
    trans[1].tx_data[0]=0;
    trans[1].tx_data[1]=x0;
    trans[1].tx_data[2]=x1>>8;
    trans[1].tx_data[3]=x1&0xff;
    trans[2].tx_data[0]=0x2B;
    trans[3].tx_data[0]=y0>>8;
    trans[3].tx_data[1]=y0&0xff;
    trans[3].tx_data[2]=y1>>8;
    trans[3].tx_data[3]=y1&0xff;
    trans[4].tx_data[0]=0x2C;
    for(int x=0; x<5; x++) spi_device_queue_trans(_lcd_spi, &trans[x], portMAX_DELAY);
    wait_trans(5);
}

void wb32_fillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color) 
{
    uint16_t line[256];
    int i, j, s;
    if((x >= 240) || (y >= 320)) return;
    if((x+w-1) >= 240)  w = 240-x;
    if((y+h-1) >= 320) h = 320-y;
    wb32_setAddrWindow(x, y, x+w-1, y+h-1);
    if (w*h < 64) {
        line[0]=color;
        for(i=0; i<w*h; i++) lcd_data((uint8_t *)line, 2);
    } else {
        if ((w*h) > 256) {
            s = 256;
            for(i=0; i<s; i++) line[i]=color;
            i=w*h;
            while(i>s){ lcd_data((uint8_t *)line, s*2); i-=s; }
            for(j=0; j<i; j++) lcd_data((uint8_t *)line, 2);
        } else {
            s = w*h;
            for(i=0; i<s; i++) line[i]=color;
            lcd_data((uint8_t *)line, s*2);
        }
    }
}

void wb32_fillScreen(uint16_t c) 
{ 
    uint16_t line[960], x, y;
    for (x=0; x<6; x++) {
        memset(&trans[x], 0, sizeof(spi_transaction_t));
        if ((x&1)==0) {
            trans[x].length=8;
            trans[x].user=(void*)0;
        } else {
            trans[x].length=8*4;
            trans[x].user=(void*)1;
        }
        trans[x].flags=SPI_TRANS_USE_TXDATA;
    }
    trans[0].tx_data[0]=0x2A;
    trans[1].tx_data[0]=0;
    trans[1].tx_data[1]=0;
    trans[1].tx_data[2]=(240)>>8;
    trans[1].tx_data[3]=(240)&0xff;
    trans[2].tx_data[0]=0x2B;
    trans[3].tx_data[0]=0;
    trans[3].tx_data[1]=0;
    trans[3].tx_data[2]=(320)>>8;
    trans[3].tx_data[3]=(320)&0xff;
    trans[4].tx_data[0]=0x2C;
    for(x=0; x<5; x++) spi_device_queue_trans(_lcd_spi, &trans[x], portMAX_DELAY);
    wait_trans(5);
    for (x=0; x<960; x++) line[x] = c;
    for(y=0; y<80; y++) lcd_data((uint8_t *)line, 1920);
}

void wb32_drawPixel(uint16_t x, uint16_t y, uint16_t color)
{
    memset(&trans[0], 0, 6*sizeof(spi_transaction_t));
    for(int x=0; x<6; x++) {
        if ((x&1)==0) {
          trans[x].length=8;
          trans[x].user=(void*)0;
        } else {
          trans[x].length=8*4;
          trans[x].user=(void*)1;
        }
        trans[x].flags=SPI_TRANS_USE_TXDATA;
    }
    trans[0].tx_data[0]=0x2A;          
    trans[1].tx_data[0]=0;
    trans[1].tx_data[1]=x;
    trans[1].tx_data[2]=0;
    trans[1].tx_data[3]=239;
    trans[2].tx_data[0]=0x2B;
    trans[3].tx_data[0]=y>>8;
    trans[3].tx_data[1]=y&0xff;
    trans[3].tx_data[2]=1;
    trans[3].tx_data[3]=63;
    trans[4].tx_data[0]=0x2C;
    trans[5].tx_data[0]=color&0xff;
    trans[5].tx_data[1]=color>>8;
    for(int x=0; x<6; x++) spi_device_queue_trans(_lcd_spi, &trans[x], portMAX_DELAY);
    wait_trans(6);
}

void wb32_pushColor(uint16_t color) 
{
    uint8_t data[2];
    data[1]=color>>8;
    data[0]=color&0xff;
    lcd_data(data,2);
}

void wb32_drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color) 
{
    uint16_t data[320];
    if((x >= 240) || (y >= 320)) return;
    if((y+h-1) >= 320) h = 320-y;
    wb32_setAddrWindow(x, y, x, y+h-1);
    for(int i=0; i<h; i++) data[i]=color;
    lcd_data((uint8_t *)data,h*2);
}

void wb32_drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color) 
{
    uint16_t data[240];
    if((x >= 240) || (y >= 320)) return;
    if((x+w-1) >= 240)  w = 240-x;
    wb32_setAddrWindow(x, y, x+w-1, y);
    for(int i=0; i<w; i++) data[i]=color;
    lcd_data((uint8_t *)data,w*2);
}

#define abs(x) ((x)<0 ? -(x) : (x))
#define swap(a, b) { int16_t t = a; a = b; b = t; }

void wb32_drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1, uint16_t color, int16_t w) 
{
  int16_t steep = abs(y1 - y0) > abs(x1 - x0);
  if (steep) { swap(x0, y0); swap(x1, y1); }
  if (x0 > x1) { swap(x0, x1); swap(y0, y1); }
  int16_t dx, dy;
  dx = x1 - x0; dy = abs(y1 - y0);
  int16_t err = dx / 2;
  int16_t ystep;
  if (y0 < y1) ystep = 1; else ystep = -1;
  for (; x0<=x1; x0++) {
    if (steep) wb32_fillRect(y0, x0, w, w, color); else wb32_fillRect(x0, y0, w, w, color);
    //if (steep) drawPixel(y0, x0, color); else drawPixel(x0, y0, color);
    err -= dy;
    if (err < 0) { y0 += ystep; err += dx; }
  }
}

void wb32_drawCircle(int16_t x0, int16_t y0, int16_t r, uint16_t color, uint16_t w) 
{
    int16_t f = 1 - r;
    int16_t ddF_x = 1;
    int16_t ddF_y = -2 * r;
    int16_t x = 0, y = r;
    wb32_fillRect(x0, y0+r, w, w, color);
    wb32_fillRect(x0  , y0-r, w, w, color);
    wb32_fillRect(x0+r, y0  , w, w, color);
    wb32_fillRect(x0-r, y0  , w, w, color);
    while (x<y) {
        if (f >= 0) { y--; ddF_y += 2; f += ddF_y; }
        x++; ddF_x += 2; f += ddF_x;
        wb32_fillRect(x0 + x, y0 + y, w, w, color); wb32_fillRect(x0 + x, y0 - y, w, w, color);
        wb32_fillRect(x0 + y, y0 + x, w, w, color); wb32_fillRect(x0 + y, y0 - x, w, w, color);
        wb32_fillRect(x0 - y, y0 - x, w, w, color); wb32_fillRect(x0 - x, y0 + y, w, w, color);
        wb32_fillRect(x0 - x, y0 - y, w, w, color); wb32_fillRect(x0 - y, y0 + x, w, w, color);
    }
}

void wb32_drawCorner(int16_t x0, int16_t y0, int16_t r, uint8_t corner, uint16_t color, uint16_t w) 
{
    int16_t f = 1 - r;
    int16_t ddF_x = 1;
    int16_t ddF_y = -2 * r;
    int16_t x = 0, y = r;
    while (x<y) {
        if (f >= 0) { y--; ddF_y += 2; f += ddF_y; }
        x++; ddF_x += 2; f += ddF_x;
        if (corner & 0x04) { 
            wb32_fillRect(x0 + x, y0 + y, w, w, color); 
            wb32_fillRect(x0 + y, y0 + x, w, w, color); 
        }
        if (corner & 0x02) {
            wb32_fillRect(x0 + x, y0 - y, w, w, color); 
            wb32_fillRect(x0 + y, y0 - x, w, w, color);
        }
        if (corner & 0x08) {
            wb32_fillRect(x0 - y, y0 + x, w, w, color); 
            wb32_fillRect(x0 - x, y0 + y, w, w, color);
        }
        if (corner & 0x01) {
            wb32_fillRect(x0 - y, y0 - x, w, w, color); 
            wb32_fillRect(x0 - x, y0 - y, w, w, color);
        }
    }
}

void wb32_fillCircle2(int16_t x0, int16_t y0, int16_t r, uint16_t color) 
{
    wb32_drawFastVLine(x0, y0-r, 2*r+1, color);
    wb32_fillCircle(x0, y0, r, 3, 0, color);
}

void wb32_fillCircle(int16_t x0, int16_t y0, int16_t r, uint8_t cornername, int16_t delta, uint16_t color) 
{
    int16_t f     = 1 - r;
    int16_t ddF_x = 1;
    int16_t ddF_y = -2 * r;
    int16_t x     = 0;
    int16_t y     = r;
    while (x<y) {
        if (f >= 0) {
            y--;
            ddF_y += 2;
            f     += ddF_y;
        }
        x++;
        ddF_x += 2;
        f     += ddF_x;

        if (cornername & 0x1) {
            wb32_drawFastVLine(x0+x, y0-y, 2*y+1+delta, color);
            wb32_drawFastVLine(x0+y, y0-x, 2*x+1+delta, color);
        }
        if (cornername & 0x2) {
            wb32_drawFastVLine(x0-x, y0-y, 2*y+1+delta, color);
            wb32_drawFastVLine(x0-y, y0-x, 2*x+1+delta, color);
        }
    }
}

void wb32_drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color) 
{
    wb32_drawFastHLine(x, y, w, color);
    wb32_drawFastHLine(x, y+h-1, w, color);
    wb32_drawFastVLine(x, y, h, color);
    wb32_drawFastVLine(x+w-1, y, h, color);
}

void wb32_drawImage(uint16_t x, uint16_t y, uint16_t w, uint16_t h, uint16_t *data)
{
    int c, n;
    wb32_setAddrWindow(x, y, x+w-1, y+h-1);
    c = w*h;
    if (c>64) {
        n=0;
        while (c > 64) {
            lcd_data((uint8_t *)(data+n),128);
            n+=64;
            c-=64;
        }
        lcd_data((uint8_t *)(data+n), c);
    } else {
        lcd_data((uint8_t *)(data), c);
    }
}

uint16_t textcolor, textbgcolor, mag=1;

int wb32_drawChar(uint16_t uniCode, uint16_t x, uint16_t y, uint8_t size, uint8_t mag)
{
    if (size) uniCode -= 32;

    uint16_t width = 0;
    uint16_t height = 0;
    const uint8_t *flash_address = 0;
    int8_t gap = 0;

    if (size == 1) {
        flash_address = chrtbl_f7[uniCode];
        width = *(widtbl_f7+uniCode);
        height = chr_hgt_f7;
        gap = 0;
    } else if (size == 2) {
        flash_address = chrtbl_f10[uniCode];
        width = *(widtbl_f10+uniCode);
        height = chr_hgt_f10;
        gap = 1;
    }

    uint16_t w = (width+ 7) / 8;
    uint16_t pX      = 0;
    uint16_t pY      = y;
    //uint16_t color   = 0;
    uint8_t line = 0;
    uint16_t i, j, k;

    if (mag==1) {
      for(i=0; i<height; i++) {
        if (textcolor != textbgcolor) wb32_drawFastHLine(x, pY, width+gap, textbgcolor);
        for (k = 0; k < w; k++) {
            line = *(flash_address + w*i + k);
            if(line) {
                pX = x + k*8;
                if(line & 0x80) wb32_drawPixel(pX, pY, textcolor);
                if(line & 0x40) wb32_drawPixel(pX+1, pY, textcolor);
                if(line & 0x20) wb32_drawPixel(pX+2, pY, textcolor);
                if(line & 0x10) wb32_drawPixel(pX+3, pY, textcolor);
                if(line & 0x8) wb32_drawPixel(pX+4, pY, textcolor);
                if(line & 0x4) wb32_drawPixel(pX+5, pY, textcolor);
                if(line & 0x2) wb32_drawPixel(pX+6, pY, textcolor);
                if(line & 0x1) wb32_drawPixel(pX+7, pY, textcolor);
            }
        }
        pY++;
      }
    } else if (mag>=2) {
      for(i=0; i<height; i++) {
        for(j=0; j<mag; j++) {
          if (textcolor != textbgcolor) wb32_drawFastHLine(x, pY, mag*(width+gap), textbgcolor);
          for (k = 0; k < w; k++) {
            line = *(flash_address + w*i + k);
            if(line) {
                pX = x + k*8;
                if(line & 0x80) wb32_drawFastHLine(pX, pY, mag,textcolor);
                if(line & 0x40) wb32_drawFastHLine(pX+1*mag, pY, mag,textcolor);
                if(line & 0x20) wb32_drawFastHLine(pX+2*mag, pY, mag,textcolor);
                if(line & 0x10) wb32_drawFastHLine(pX+3*mag, pY, mag,textcolor);
                if(line & 0x8) wb32_drawFastHLine(pX+4*mag, pY, mag,textcolor);
                if(line & 0x4) wb32_drawFastHLine(pX+5*mag, pY, mag,textcolor);
                if(line & 0x2) wb32_drawFastHLine(pX+6*mag, pY, mag,textcolor);
                if(line & 0x1) wb32_drawFastHLine(pX+7*mag, pY, mag,textcolor);
            }
          }
          pY++;
        }
      }
    }
    return (width+gap)*mag;
}

int wb32_drawString(const char *string, uint16_t poX, uint16_t poY, uint8_t size, uint8_t mag)
{
    uint16_t sumX = 0;
    while(*string) {
        uint16_t xPlus = wb32_drawChar(*string, poX, poY, size, mag);
        sumX += xPlus;
        string++;
        poX += xPlus;
    }
    return sumX;
}

void wb32_setTextColor(uint16_t c1, uint16_t c2)
{
    textcolor = c1;
    textbgcolor = c2;
}

void lcdInit()
{
    spi_bus_config_t buscfg={
        .miso_io_num=PIN_NUM_MISO, .mosi_io_num=PIN_NUM_MOSI, .sclk_io_num=PIN_NUM_CLK,
        .quadwp_io_num=-1, .quadhd_io_num=-1
    };
    spi_device_interface_config_t devcfg={
        .clock_speed_hz=40000000, .mode=0, .spics_io_num=PIN_NUM_CS,
        .queue_size=7, .pre_cb=lcd_spi_pre_transfer_callback,
    };
    spi_bus_initialize(HSPI_HOST, &buscfg, 1);
    spi_bus_add_device(HSPI_HOST, &devcfg, &_lcd_spi);
    lcd_init(_lcd_spi);
    //display_pretty_colors(spi);
}

void wb32_init()
{
	lcdInit();
	wb32_fillScreen(0);
	REG_CLR_BIT(RTC_IO_XTAL_32K_PAD_REG, RTC_IO_X32P_MUX_SEL); /* gpio32 route to digital io_mux */
    REG_CLR_BIT(RTC_IO_XTAL_32K_PAD_REG, RTC_IO_X32N_MUX_SEL); /* gpio33 route to digital io_mux */
}

////////////// WB32 Sprite Engine //////////////////

#define U16x2toU32(m,l) ((((uint32_t)(l>>8|(l&0xFF)<<8))<<16)|(m>>8|(m&0xFF)<<8))
uint8_t *wb32_buf8; // WiFiBoy32 8bit-mode offscreen buffer !! test for 3d-engine earth demo 128x128 16bit
uint16_t wb32_pal8[256];

uint16_t swapbyte(uint16_t x) { 
  return (x>>8) | (x<<8);
}

uint16_t wb32_color565(uint8_t r, uint8_t g, uint8_t b) 
{
  return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3);
}

void wb32_setPal8(uint8_t n, uint16_t c)
{
    wb32_pal8[n]=(c>>8)|(c<<8);
}

void wb32_blit8()
{
	uint16_t line[960], x, y;
    for (x=0; x<6; x++) {
        memset(&trans[x], 0, sizeof(spi_transaction_t));
        if ((x&1)==0) {
            trans[x].length=8;
            trans[x].user=(void*)0;
        } else {
            trans[x].length=8*4;
            trans[x].user=(void*)1;
        }
        trans[x].flags=SPI_TRANS_USE_TXDATA;
    }
    trans[0].tx_data[0]=0x2A;          //Column Address Set
    trans[1].tx_data[0]=0;             //Start Col High
    trans[1].tx_data[1]=0;             //Start Col Low
    trans[1].tx_data[2]=(240)>>8;      //End Col High
    trans[1].tx_data[3]=(240)&0xff;    //End Col Low
    trans[2].tx_data[0]=0x2B;          //Page address set
    trans[3].tx_data[0]=0;             //Start page high
    trans[3].tx_data[1]=0;             //start page low
    trans[3].tx_data[2]=(320)>>8;      //end page high
    trans[3].tx_data[3]=(320)&0xff;    //end page low
    trans[4].tx_data[0]=0x2C;          //memory write
    for(x=0; x<5; x++) spi_device_queue_trans(_lcd_spi, &trans[x], portMAX_DELAY);
    wait_trans(5);
    for(y=0; y<80; y++) {
		for (x=0; x<960; x++) {
			line[x] = wb32_pal8[wb32_buf8[x+y*960]]; 
		}
		lcd_data((uint8_t *)line, 1920);
	}
}

void wb32_blitBuf8(uint16_t xs, uint16_t ys, uint16_t ws, int xd, int yd, uint16_t width, uint16_t height, uint8_t *data)
{
    uint8_t d;
    int dx, dy;
    for (int y=0; y<height; y++) {
        for(int x=0; x<width; x++) {
            dx=x+xd;
            dy=y+yd;
            if (dx>=0 && dx<240 && dy>=0 && dy<320) {
                d = data[(y+ys)*ws+x+xs];
                if (d!=0) wb32_buf8[dx+(dy)*240] = d;
            }
        }
    }
}

void wb32_rot8(uint16_t dx, uint16_t dy, double angle, double scale, uint16_t offx, uint16_t offy, uint8_t w, uint8_t h, uint8_t *sprite)
{
    double duCol, dvCol, duRow, dvRow, startingu, startingv;
    double fSrcRotCX, fSrcRotCY, fDstRotCX, fDstRotCY;
    double rowu, rowv, u, v;
    int x, y, height, width, hw, hh;
    uint8_t d;
    
    if (w>h) {
        height=w*(1.5*scale);
        width=w*(1.5*scale);
    } else {
        height=h*(1.5*scale);
        width=h*(1.5*scale);
    }
    fSrcRotCX=(double)w/2.0;
    fSrcRotCY=(double)h/2.0;
    fDstRotCX=(double)(width/2); 
    fDstRotCY=(double)(height/2);
    
    duCol = sin(-angle) * (1.0 / scale);
    dvCol = cos(-angle) * (1.0 / scale);
    duRow = dvCol;
    dvRow =-duCol;
    startingu = fSrcRotCX -(fDstRotCX * dvCol + fDstRotCY * duCol);
    startingv = fSrcRotCY -(fDstRotCX * dvRow + fDstRotCY * duRow);
    
    rowu = startingu;
    rowv = startingv;
    hw = width / 2;
    hh = height / 2;
    for(y=0; y<height; y++) {
        u = rowu;
        v = rowv;
        for(x=0; x<width; x++) {
            if(u>0 && v>0 && u<w && v<h) {
                d = sprite[(int)(u)+offx+((int)(v)+offy)*240];
                if (d>0) wb32_buf8[x+dx-hw+(y+dy-hh)*240] = d;
            }
            //else wb_buf8[x+dx+(y+dy)*240] = 0;
            u += duRow;
            v += dvRow;
        }
        rowu += duCol;
        rowv += dvCol;
    }
}

void wb32_setBuf8(uint32_t i, uint8_t d)
{
    wb32_buf8[i]=d;
    //wb32_buf8[i+1]=d;
}

void wb32_clearBuf8()
{
	memset(wb32_buf8, 0, 76800);
}

void wb32_initBuf8()
{
	wb32_buf8=(uint8_t *)malloc(76800);
}

/////////////////////////////////

// Ticker Library

#include "esp32-hal-timer.h"

hw_timer_t * wb32_timer = NULL;  // the esp32 timer

void wb32_tickerInit(uint32_t us, void code())
{
  wb32_timer = timerBegin(0, 80, true); // hw timer #0, divider=80 (1us)
  timerAttachInterrupt(wb32_timer, code, true); // attach service routine, true: edge
  timerAlarmWrite(wb32_timer, us, true); // timer=500000us=500ms, true=auto-reload
  timerAlarmEnable(wb32_timer); // eanble timer#0
}

void wb32_tickerAlarm(uint32_t us)
{
  timerAlarmWrite(wb32_timer, us, true);
}

void wb32_tickerEnable()
{
  timerAlarmEnable(wb32_timer);
}

void wb32_tickerDisable()
{
  timerAlarmEnable(wb32_timer);
}
